#include <stdio.h>
#include <stdlib.h>
#include "fxlib.h"
#include "MonochromeLib.h"
#include "fonctions basiques.h"
#include "Ether.h"

char mouvement_a_effectuer(Map *map,Etre *etre)
{
	//map : structure qui contient des données concernant la carte
	/*etre : structure correspondant à l'entité qui se déplace, avec notammment :
		x et y : coordonnées de l'être
		cible_x et cible_y : coordonnées du point de destination
	
	*/
	char i,j,x,y;
	Noeud *noeud_courant=nouveau_noeud(etre->x,etre->y,NULL),*noeud,*meilleur_noeud;
	Liste *liste_ouverte=initialisation(),*liste_fermee=initialisation();		//On initialise lla liste ouverte et la liste fermee
	noeud_courant=insertion(noeud_courant,liste_fermee);		//Le tout premier noeud est dans la liste fermee
	
	if(confine(etre->x,etre->y,map) || confine(etre->cible_x,etre->cible_y,map) || (etre->x==etre->cible_x && etre->y==etre->cible_y))
	{
		suppression_liste(liste_ouverte);
		suppression_liste(liste_fermee);
		return 0;
	}
	//Pour éviter tout calcul inutile, on retourne 0 si la cible ou l'objet sont coincés, où si l'objet est déjà sur la cible
	
	while(noeud_courant->x!=etre->cible_x || noeud_courant->y!=etre->cible_y)
	{
		//On examine les cases adjacentes :
		for(i=-1;i<=1;i++)
		{
			for(j=-1;j<=1;j++)
			{
				x=noeud_courant->x+i;
				y=noeud_courant->y+j;
				if((MAP(x,y)<10 || MAP(x,y)==REMPART || (x==etre->cible_x && y==etre->cible_y)) && (i+j)&1)		//On exclut les obstacles ou les cases en diagonales
				{
					noeud=nouveau_noeud(x,y,noeud_courant);		//On déclare un nouveau noeud
					if(!recherche_liste(noeud,liste_fermee))	//S'il n'est pas dans la liste fermée, on le met en liste ouverte
					{
						noeud->distance=DISTANCE(x,y,etre->cible_x,etre->cible_y)+50*(MAP(x,y)==REMPART);		//La distance à la cible détermine la qualité du noeud
						noeud=insertion(noeud,liste_ouverte);
					}
					else free(noeud);		//Sinon on l'oublie
				}
			}
		}
		
		
		//Recherche du meilleur noeud :
		meilleur_noeud=liste_ouverte->premier;
		noeud=meilleur_noeud;
		if(!noeud)		//Si la liste ouverte est vide, on quitte
		{
			suppression_liste(liste_ouverte);
			suppression_liste(liste_fermee);
			return 0;		
		}
		
		//On trouve le meilleur noeud en fonction de la distance :
		while(noeud!=NULL)
		{
			if(noeud->distance < meilleur_noeud->distance)meilleur_noeud=noeud;
			noeud=noeud->suivant;
		}
		
		//On exclut le meilleur noeud de la liste ouverte :
		if(liste_ouverte->premier)
		{
			if(meilleur_noeud==liste_ouverte->premier)
				liste_ouverte->premier=meilleur_noeud->suivant;
			else meilleur_noeud->precedent->suivant=meilleur_noeud->suivant;
			
			if(meilleur_noeud==liste_ouverte->dernier)
				liste_ouverte->dernier=meilleur_noeud->precedent;
			else meilleur_noeud->suivant->precedent=meilleur_noeud->precedent;
			
			liste_ouverte->taille--;
		}
		
		//On l'insère dans la liste fermée :
		meilleur_noeud=insertion(meilleur_noeud,liste_fermee);

		//On prend le meilleur noeud comme noeud courant :
		noeud_courant=meilleur_noeud;
	}
	//On quitte la boucle lorsque le noeud courant correspond à la destination recherchée
	
	noeud=noeud_courant;
	while(noeud->parent->parent)noeud=noeud->parent;		//On cherche le second noeud du trajet
	
	if(noeud->y < noeud->parent->y)i=1;
	if(noeud->x > noeud->parent->x)i=2;
	if(noeud->y > noeud->parent->y)i=3;
	if(noeud->x < noeud->parent->x)i=4;
	//On en déduit le mouvement à effectuer
	
	suppression_liste(liste_ouverte);
	suppression_liste(liste_fermee);
	
	return i;
}

Noeud *nouveau_noeud(char x,char y,Noeud *parent)
{
	//Crée un noeud et en retourne l'adresse mémoire
	Noeud *nouveau=malloc(sizeof(Noeud));
	nouveau->x=x;
	nouveau->y=y;
	nouveau->distance=0;
	nouveau->parent=parent;
	
	return nouveau;
}

Noeud *insertion(Noeud *noeud,Liste *liste)
{
	/*Introduit un noeud dans une liste. Si le noeud y est déjà présent, on en retourne l'adresse mémoire et on supprime l'autre (pour éviter
	  les doublons)*/

	Noeud *element=liste->premier;
	while(element!=NULL)
	{
		if(element->x==noeud->x && element->y==noeud->y)	//Si le noeud est déjà dans la liste, on se contente de le mettre à jour
		{
			element->parent=noeud->parent;
			free(noeud);
			return element;
		}
		
		element=element->suivant;
	}


	if(liste->taille)
	{
		liste->dernier->suivant=noeud;
		noeud->precedent=liste->dernier;
	}
	else
	{
		liste->premier=noeud;
		noeud->precedent=NULL;
	}
	
	liste->dernier=noeud;
	noeud->suivant=NULL;
	liste->taille++;
	
	return noeud;
}

Bool recherche_liste(Noeud *noeud,Liste *liste)
{
	//Determine si un noeud se situe ou non dans une liste

	Noeud *element=liste->premier;
	while(element!=NULL)
	{
		if(element->x==noeud->x && element->y==noeud->y)return VRAI;
		element=element->suivant;
	}

	return FAUX;
}

Liste *initialisation()
{
	//Crée une liste et retourne un pointeur
	Liste *liste=malloc(sizeof(Liste));
    liste->premier=NULL;
	liste->dernier=NULL;
	liste->taille=0;
	return liste;
}

void suppression_liste(Liste *liste)
{
	//Détruit une liste ainsi que tous les éléments qu'elle contient :
	Noeud *element=liste->premier;
	Noeud *element_suivant;
	
	while(element)
	{
		element_suivant=element->suivant;
		free(element);
		element=element_suivant;
	}

	free(liste);
}